/*
 *  Interactive disassembler (IDA).
 *  TheHyper VM Processor Module.
 *
 *  (c) 2006 Rolf Rolles
 */

#include "reg.hpp"
#include "out.hpp"
#include "ana.hpp"
#include "emu.hpp"
#include "ins.hpp"
#include "TH.hpp"

#include <map>

/* Functions for dealing with this VM's keyed memory regions */
static std::map<unsigned long, ea_t> keyed_regions;

ea_t get_keyed_address(unsigned long key) { return keyed_regions[key]; };

void find_keyed_regions()
{
    ea_t retval = inf.minEA;
    ea_t endval = inf.maxEA;
    msg("retval = %lx, endval = %lx\n", retval, endval);
    while( ( retval = bin_search          // find all occurrences of the VM
                                ( retval, // memory delimiting keyword, $$$$
                                  endval,
                                  "$$$$", 
                                    NULL, 
                                       4, 
                      BIN_SEARCH_FORWARD, 
                      BIN_SEARCH_NOCASE ) 
           ) != BADADDR )
    {
        doDwrd(retval, 8);
        ea_t key = get_long(retval + 4);
        retval = retval + 8;
        keyed_regions[key] = retval;
    }
}

// Register number from width and number of register
TH_regs TH_Regnum(unsigned long width, unsigned long regnum)
{
    return (TH_regs)( (TH_regs)(regnum * 3) + (TH_regs)(width >> 1) );
}

static asm_t thasm = {
  AS_COLON | ASH_HEXF0 | ASD_DECF0 | ASO_OCTF0 | ASB_BINF0 | AS_N2CHR | AS_ONEDUP,
  0,
  "TheHyper VM Imaginary Assembler",
  0,
  NULL,     //header
  NULL,
  ".org",
  ".end",

  ";",        // comment string
  '"',        // string delimiter
  '\'',       // char delimiter
  "'\"",      // special symbols in char and string constants

  ".db",    // ascii string directive
  ".db",    // byte directive
  ".dw",    // word directive
  ".dd",     // no double words
  ".dq",     // no qwords
  NULL,     // oword  (16 bytes)
  NULL,     // no float
  NULL,     // no double
  NULL,     // no tbytes
  NULL,     // no packreal
  NULL,     //".db.#s(b,w) #d,#v",   // #h - header(.byte,.word)
                    // #d - size of array
                    // #v - value of array elements
                    // #s - size specifier
  ".rs %s",    // uninited data (reserve space)
  ".equ",
  NULL,         // seg prefix
  NULL,         // preline for checkarg
  NULL,      // checkarg_atomprefix
  NULL,   // checkarg operations
  NULL,   // XlatAsciiOutput
  "*",    // a_curip
  NULL,         // returns function header line
  NULL,         // returns function footer line
  NULL,         // public
  NULL,         // weak
  NULL,         // extrn
  NULL,         // comm
  NULL,         // get_type_name
  NULL,         // align
  '(', ')',     // lbrace, rbrace
  NULL,    // mod
  NULL,    // and
  NULL,    // or
  NULL,    // xor
  NULL,    // not
  NULL,    // shl
  NULL,    // shr
  NULL,    // sizeof
};

static char *shnames[] = { "th", NULL };
static char  *lnames[] = { "TheHyper VM", NULL };
static asm_t *  asms[] = { &thasm, NULL };

static uchar retcTH[] = { 0x23 };    //retd
static bytes_t retcodes[] = { { sizeof(retcTH), retcTH }, { 0, NULL } };

static int notify(processor_t::idp_notify msgid, ...)
{
  va_list va;
  va_start(va, msgid);

// A well behaving processor module should call invoke_callbacks()
// in his notify() function. If this function returns 0, then
// the processor module should process the notification itself
// Otherwise the code should be returned to the caller:

  static bool bFoundRegions = 0;
  if(!bFoundRegions && inf.minEA != BADADDR)
  {
    find_keyed_regions();
    bFoundRegions = 1;
  }

  int code = invoke_callbacks(HT_IDP, msgid, va);
  if ( code ) return code;

  switch(msgid)
  {
    case processor_t::init:
        // msg("processor_t::init notification called\n");
        inf.mf = 0;
        break;

    case processor_t::newfile:
    case processor_t::oldfile:
        find_keyed_regions();
        // msg("processor_t::newfile notification called\n");
        break;

    case processor_t::newprc:
        // msg("processor_t::newprc notification called\n");
        break;

    case processor_t::newseg:    // new segment
        msg("processor_t::newseg notification called\n");
        break;

  }
  va_end(va);

  return(1);
}

//-----------------------------------------------------------------------
//      Processor Definition
//-----------------------------------------------------------------------
processor_t LPH =
{
  IDP_INTERFACE_VERSION,
  PLFM_TH,              // id
  PRN_HEX|PR_SEGTRANS,
  8,                    // 8 bits in a byte for code segments
  8,                    // 8 bits in a byte for other segments

  shnames,
  lnames,

  asms,

  notify,

  header,
  footer,

  segstart,
  std_gen_segm_footer,

  NULL,

  ana,
  emu,

  out,
  outop,
  intel_data,
  NULL,         //  cmp_opnd,  // 0 if not cmp 1 if eq
  NULL,         //  can_have_type,  //&op    // 1 -yes 0-no    //reg

  qnumber(TH_regnames),            // Number of registers
  TH_regnames,                     // Regsiter names
  NULL,                         // get abstract register

  0,                            // Number of register files
  NULL,                         // Register file names
  NULL,                         // Register descriptions
  NULL,                         // Pointer to CPU registers

  rVcs,rVds,
  0,                            // size of a segment register
  rVcs,rVds,

  NULL,                         // No known code start sequences
  retcodes,

  0,
  TH_end,
  Instructions,
  NULL,                 // int  (*is_far_jump)(int icode);
  NULL,                 // Translation function for offsets
  0,                    // int tbyte_size;  -- doesn't exist
  NULL,                 // int (*realcvt)(void *m, ushort *e, ushort swt);
  { 0, 0, 0, 0 },       // char real_width[4];
                        // number of symbols after decimal point
                        // 2byte float (0-does not exist)
                        // normal float
                        // normal double
                        // long double
  NULL,                 // int (*is_switch)(switch_info_t *si);
  NULL,                 // long (*gen_map_file)(FILE *fp);
  NULL,                 // ulong (*extract_address)(ulong ea,const char *string,int x);
  NULL,                 // int (*is_sp_based)(op_t &x); -- always, so leave it NULL
  NULL,                 // int (*create_func_frame)(func_t *pfn);
  NULL,                 // int (*get_frame_retsize(func_t *pfn)
  NULL,                 // void (*gen_stkvar_def)(char *buf,const member_t *mptr,long v);
  gen_spcdef,           // Generate text representation of an item in a special segment

};
